(**
   Хȥåƥ ץ륳
   Х
   
   $Id: bug.ml,v 1.1 2007/05/23 04:20:10 ogasawara Exp $
*)
(*
   Copyright (c) 2007 ogasawara@itpl.co.jp All Rights Reserved.
   
   Permission is hereby granted, free of charge, to any person obtaining
   a copy of this software and associated documentation files (the
   "Software"), to deal in the Software without restriction, including
   without limitation the rights to use, copy, modify, merge, publish,
   distribute, sublicense, and/or sell copies of the Software, and to
   permit persons to whom the Software is furnished to do so, subject to
   the following conditions:
  
   The above copyright notice and this permission notice shall be
   included in all copies or substantial portions of the Software.

   THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,   
   EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
   MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
   IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY
   CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
   TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
   SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
*)

open Printf
open Base
open Mysql

(* Хơ *)
type status = 
    REPORTED  (*  *)
  | OPEN      (* Ĵ *)
  | PENDING   (*  *)
  | CLOSE     (* λ *)

let status_list = [REPORTED; OPEN; PENDING; CLOSE]

let status2raw = function
    REPORTED -> 0
  | OPEN -> 1
  | PENDING -> 2
  | CLOSE -> 3

exception INVALID_STATUS_CODE of int

let raw2status = function
    0 -> REPORTED
  | 1 -> OPEN
  | 2 -> PENDING
  | 3 -> CLOSE
  | x -> raise (INVALID_STATUS_CODE x)

let string_of_status = function
    REPORTED -> ""
  | OPEN -> "Ĵ"
  | PENDING -> ""
  | CLOSE -> "λ"

(* ХФвǡ *)
type process = { 
    body_id : int;     (* process.pidб *)
    date : Calendar.t; (* Ȥ줿 *)
    comment : string   (*  *)
  }

(* Х *)
type bug = { 
    id : int;                (* ͭID *)
    title : string;          (* ȥ *)
    status : status;         (* Хơ *)
    process : process list;  (* в *)
  }

let has_row db sql =
  try
    if Mysql.size (Mysql.exec db sql) = 0L then
      false
    else
      true
  with
    _ -> false

(* bugsơ֥idĹԤ뤫? *)
let maybe_id db = function
    Some id ->
      if has_row db (sprintf "select id from bugs where id = %d" id) then
	Some id
      else
	None
  | None ->
      None

(* processơ֥idĹԤ뤫? *)
let maybe_body_id db = function
    Some id ->
      if has_row db (sprintf "select pid from process where pid = %d" id) then
	Some id
      else
	None
  | None ->
      None

(* ꤵ줿idbody_idΥǡDBФbugˤؿ *)
let take db id body_id =
  let sql = 
    sprintf  (* SQLʸ *)
      ("select A.id, A.title, A.status, B.pid, B.day, B.comment " ^^
       "from bugs A, process B " ^^
       " where A.id = B.id and " ^^
       " A.id = %d " ^^
       " order by B.day") id
  in
  (*ԼäbugѴ *)
  let build bug col =
    let build_process col =
      { body_id = not_null int2ml 
                  @@ col ~key:"pid";
	date = Calendar.of_string 
	       @@ not_null str2ml 
               @@ col ~key:"day";
	comment = not_null str2ml @@ col ~key:"comment" }
    in
    if bug.id = -1 then
      { id = not_null int2ml 
             @@ col ~key:"id";
	title = not_null str2ml 
                @@ col ~key:"title";
	status = raw2status 
	         @@ not_null int2ml 
                 @@ col ~key:"status";
	process = [ build_process col ] }
    else
      { bug with process = (build_process col) :: bug.process }
  in
  (* SQLʸμ¹Ԥbugι *)
  let bug =
    Db.fold db sql build 
      { id = -1;
	title = "";
	status = OPEN;
	process = [] }
  in
  let bug' =
    { bug with process = List.rev bug.process }
  in
  let focus_process =
    match body_id with
      None when List.length bug'.process > 0 -> 
	Some (List.hd bug'.process)
    | None -> None
    | Some body_id ->
	try
	  Some (List.find (fun p -> p.body_id = body_id) bug'.process)
	with
	  Not_found -> None
  in
  bug', focus_process

(* ȥȥơ򸡺ˡDBbug listؿ *)
let list db title status =
  let where =
    let c1 = (* ȥlikeˤʬ *)
      match title with
	Some v -> Some (sprintf " (title like '%%%s%%') " (escape v))
      | None -> None
    in
    let c2 = (* ơˤʤꤳ *)
      match status with
	[] -> None
      | _ -> Some (string_of_list 
		     (sprintf " status = %d " $ status2raw) ("("," or ", ")") status)
    in
    match c1, c2 with
      Some x, Some y -> sprintf "and (%s and %s)" x y
    | Some x, None -> sprintf "and %s" x
    | None, Some y -> sprintf "and %s" y
    | _ -> ""
  in
  let sql = 
    sprintf  (* SQLʸ *)
      ("select A.id, A.title, A.status, B.pid, B.day " ^^
       "from bugs A, process B " ^^
       "where A.id = B.id " ^^
       " and B.day >= " ^^
       "(select max(C.day) from process C where C.id = A.id)" ^^
       " %s " ^^
       "order by B.day") where 
  in
  (*ԼäbugѴ *)
  let build col = { id = not_null int2ml 
		         @@ col ~key:"id";
		    title = not_null str2ml 
                            @@ col ~key:"title";
		    status = raw2status 
			     @@ not_null int2ml 
                             @@ col ~key:"status";
		    process = [ { body_id = not_null int2ml
                                            @@ col ~key:"pid";
				  date = Calendar.of_string 
			                 @@ not_null str2ml 
                                         @@ col ~key:"day";
				  comment = "" } ] }
  in
  (* SQLʸμ¹Ԥbug listι *)
  List.rev (Db.map db sql build)

